==> oidc-1.0.0.tar.gz: oidc/src/Controller/OpenidConnectController.php <==
<?php

namespace Drupal\oidc\Controller;

use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\SessionManagerInterface;
use Drupal\Core\Url;
use Drupal\externalauth\AuthmapInterface;
use Drupal\oidc\OpenidConnectLoginException;
use Drupal\oidc\OpenidConnectSessionInterface;
use Drupal\oidc\Routing\ImmutableTrustedRedirectResponse;
use Drupal\externalauth\ExternalAuthInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Controller for the OpenID Connect requests.
 */
class OpenidConnectController extends ControllerBase {

  /**
   * The session manager.
   *
   * @var \Drupal\Core\Session\SessionManagerInterface
   */
  protected $sessionManager;

  /**
   * The OpenID Connect session service.
   *
   * @var \Drupal\oidc\OpenidConnectSessionInterface
   */
  protected $session;

  /**
   * The external authentication service.
   *
   * @var \Drupal\externalauth\ExternalAuthInterface
   */
  protected $externalauth;

  /**
   * The authentication mapping service.
   *
   * @var \Drupal\externalauth\AuthmapInterface
   */
  protected $authmap;

  /**
   * The UUID service.
   *
   * @var \Drupal\Component\Uuid\UuidInterface
   */
  protected $uuid;

  /**
   * The user storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $userStorage;

  /**
   * The logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $loggerChannel;

  /**
   * Class constructor.
   *
   * @param \Drupal\Core\Session\SessionManagerInterface $session_manager
   *   The session manager.
   * @param \Drupal\oidc\OpenidConnectSessionInterface $session
   *   The OpenID Connect session service.
   * @param \Drupal\externalauth\ExternalAuthInterface $externalauth
   *   The external authentication service.
   * @param \Drupal\externalauth\AuthmapInterface $authmap
   *   The authentication mapping service.
   * @param \Drupal\Component\Uuid\UuidInterface
   *   The UUID service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger_channel
   *   The logger channel.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(SessionManagerInterface $session_manager, OpenidConnectSessionInterface $session, ExternalAuthInterface $externalauth, AuthmapInterface $authmap, UuidInterface $uuid, EntityTypeManagerInterface $entity_type_manager, LoggerChannelInterface $logger_channel) {
    $this->sessionManager = $session_manager;
    $this->session = $session;
    $this->externalauth = $externalauth;
    $this->authmap = $authmap;
    $this->uuid = $uuid;
    $this->userStorage = $entity_type_manager->getStorage('user');
    $this->loggerChannel = $logger_channel;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('session_manager'),
      $container->get('oidc.openid_connect_session'),
      $container->get('externalauth.externalauth'),
      $container->get('externalauth.authmap'),
      $container->get('uuid'),
      $container->get('entity_type.manager'),
      $container->get('logger.channel.oidc')
    );
  }

  /**
   * Start the OpenID Connect authentication.
   *
   * @param string $realm
   *   The realm plugin ID.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Drupal\oidc\Routing\ImmutableTrustedRedirectResponse
   *   Redirect to the authorization endpoint.
   */
  public function login($realm, Request $request) {
    // Initialize the realm.
    try {
      $this->session->initRealm($realm);
      $plugin = $this->session->getRealmPlugin();
    }
    catch (PluginNotFoundException $ex) {
      $this->session->destroy();
      throw new NotFoundHttpException();
    }

    // Ensure it's enabled.
    if (!$plugin->isEnabled()) {
      $this->session->destroy();
      throw new NotFoundHttpException();
    }

    // Initialize the state.
    $destination = $request->query->get('destination');
    $state = $this->session->initState($destination);

    // Get and redirect to the login URL.
    $redirect_url = Url::fromRoute('oidc.openid_connect.login_redirect');
    $url = $plugin->getLoginUrl($state, $redirect_url)
      ->setAbsolute();

    return new ImmutableTrustedRedirectResponse($url);
  }

  /**
   * Redirect callback triggered after an OpenID Connect login.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to the front page.
   */
  public function loginRedirect(Request $request) {
    $query = $request->query;
    $realm = $this->session->getRealmPluginId();

    if ($query->has('error') && $query->get('error_description')) {
      // An error occurred.
      $this->loggerChannel->critical('@error error occurred on @realm realm: %description', [
        '@error' => $query->get('error'),
        '@realm' => $realm,
        '%description' => $query->get('error_description'),
      ]);
    }
    elseif (!$query->has('code')) {
      // No code specified.
      $this->loggerChannel->error('Code missing in redirect from @realm realm.', [
        '@realm' => $realm,
      ]);
    }
    else {
      // Get the realm plugin.
      $plugin = $this->session->getRealmPlugin();

      try {
        // Get the tokens.
        $tokens = $plugin->getJsonWebTokens($this->session->getState(), $query->get('code'));
        $this->session->setJsonWebTokens($tokens);

        // Check for an existing user.
        $authname = $tokens->getId();
        $provider = 'oidc:' . $realm;
        $account = $this->externalauth->load($authname, $provider);

        // No user found, try to link to an existing user by email.
        if (!$account && $tokens->getEmail()) {
          $accounts = $this->userStorage->loadByProperties([
            'mail' => $tokens->getEmail(),
          ]);

          /** @var \Drupal\user\UserInterface $account */
          foreach ($accounts as $account) {
            if (!$this->authmap->getAll($account->id())) {
              $this->externalauth->linkExistingAccount($authname, $provider, $account);
              $account->setPassword(NULL)->save();
              break;
            }

            $account = NULL;
          }
        }

        // Email not set or no valid user found, create a new one.
        if (!$account) {
          do {
            $name = $this->uuid->generate();
            $count = $this->userStorage->getQuery()
              ->count()
              ->condition('name', $name)
              ->execute();
          } while ($count);

          $account = $this->externalauth->register($authname, $provider, [
            'name' => $name,
          ]);
        }

        // Login.
        $this->externalauth->userLoginFinalize($account, $authname, $provider);

        // Clear the state and restore the destination.
        if (($destination = $this->session->clearState()) !== NULL) {
          $request->query->set('destination', $destination);
        }
      }
      catch (OpenidConnectLoginException $ex) {
        if ($this->currentUser()->isAnonymous()) {
          // Redirect to the destination.
          $redirect = new ImmutableTrustedRedirectResponse($ex->getDestination());
          $this->session->destroy();
        }
        else {
          // Logout before redirecting to the destination.
          $request->query->set('destination', $ex->getDestination()->getInternalPath());
          $redirect = $this->logout($request);
        }

        // Set the exception message as error.
        if (($message = $ex->getMessage()) !== '') {
          $this->messenger()->addError($message);
        }

        return $redirect;
      }
      catch (\Exception $ex) {
        $this->loggerChannel->error('Error on authentication via @realm realm: %error', [
          '@realm' => $realm,
          '%error' => $ex->getMessage(),
        ]);
      }
    }

    // Destroy the session and set an error message if authentication failed.
    if ($this->currentUser()->isAnonymous()) {
      $this->session->destroy();
      $this->messenger()->addError($this->t('Authentication failed, please try again later.'));
    }

    return $this->redirect('<front>');
  }

  /**
   * Start the OpenID Connect logout.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Drupal\oidc\Routing\ImmutableTrustedRedirectResponse
   *   Redirect to the end session endpoint.
   */
  public function logout(Request $request) {
    // Get the realm plugin.
    $plugin = $this->session->getRealmPlugin();
    $plugin_id = $this->session->getRealmPluginId();

    // Get the tokens.
    $tokens = $this->session->getJsonWebTokens();

    // Logout.
    user_logout();

    // Regenerate the session. This is required because user_logout() destroys it,
    // but the OpenID Connect session depends on it.
    $this->sessionManager->clear();
    $this->sessionManager->regenerate();

    // Restore the realm since the session was destroyed.
    $this->session->initRealm($plugin_id);

    // Initialize the state.
    $destination = $request->query->get('destination');
    $state = $this->session->initState($destination);

    // Get and redirect to the logout URL.
    $redirect_url = Url::fromRoute('oidc.openid_connect.logout_redirect');
    $url = $plugin->getLogoutUrl($tokens, $state, $redirect_url)
      ->setAbsolute();

    return new ImmutableTrustedRedirectResponse($url);
  }

  /**
   * Redirect callback triggered after an OpenID Connect logout.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to the front page.
   */
  public function logoutRedirect(Request $request) {
    // Set a message if there are no error messages.
    if (!$this->messenger()->messagesByType(MessengerInterface::TYPE_ERROR)) {
      $this->messenger()->addStatus($this->t('You have been logged out successfully.'));
    }

    // Clear the state and restore the destination.
    if (($destination = $this->session->clearState()) !== NULL) {
      $request->query->set('destination', $destination);
    }

    $this->session->destroy();

    return $this->redirect('<front>');
  }

}

==> oidc-1.0.3.tar.gz: oidc/src/Controller/OpenidConnectController.php <==
<?php

namespace Drupal\oidc\Controller;

use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AnonymousUserSession;
use Drupal\Core\Session\SessionManagerInterface;
use Drupal\Core\Url;
use Drupal\externalauth\AuthmapInterface;
use Drupal\oidc\OpenidConnectLoginException;
use Drupal\oidc\OpenidConnectSessionInterface;
use Drupal\oidc\Routing\ImmutableTrustedRedirectResponse;
use Drupal\externalauth\ExternalAuthInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Controller for the OpenID Connect requests.
 */
class OpenidConnectController extends ControllerBase {

  /**
   * The session manager.
   *
   * @var \Drupal\Core\Session\SessionManagerInterface
   */
  protected $sessionManager;

  /**
   * The OpenID Connect session service.
   *
   * @var \Drupal\oidc\OpenidConnectSessionInterface
   */
  protected $session;

  /**
   * The external authentication service.
   *
   * @var \Drupal\externalauth\ExternalAuthInterface
   */
  protected $externalauth;

  /**
   * The authentication mapping service.
   *
   * @var \Drupal\externalauth\AuthmapInterface
   */
  protected $authmap;

  /**
   * The UUID service.
   *
   * @var \Drupal\Component\Uuid\UuidInterface
   */
  protected $uuid;

  /**
   * The user storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $userStorage;

  /**
   * Class constructor.
   *
   * @param \Drupal\Core\Session\SessionManagerInterface $session_manager
   *   The session manager.
   * @param \Drupal\oidc\OpenidConnectSessionInterface $session
   *   The OpenID Connect session service.
   * @param \Drupal\externalauth\ExternalAuthInterface $externalauth
   *   The external authentication service.
   * @param \Drupal\externalauth\AuthmapInterface $authmap
   *   The authentication mapping service.
   * @param \Drupal\Component\Uuid\UuidInterface
   *   The UUID service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(SessionManagerInterface $session_manager, OpenidConnectSessionInterface $session, ExternalAuthInterface $externalauth, AuthmapInterface $authmap, UuidInterface $uuid, EntityTypeManagerInterface $entity_type_manager) {
    $this->sessionManager = $session_manager;
    $this->session = $session;
    $this->externalauth = $externalauth;
    $this->authmap = $authmap;
    $this->uuid = $uuid;
    $this->userStorage = $entity_type_manager->getStorage('user');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('session_manager'),
      $container->get('oidc.openid_connect_session'),
      $container->get('externalauth.externalauth'),
      $container->get('externalauth.authmap'),
      $container->get('uuid'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * Start the OpenID Connect authentication.
   *
   * @param string $realm
   *   The realm plugin ID.
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Drupal\oidc\Routing\ImmutableTrustedRedirectResponse
   *   Redirect to the authorization endpoint.
   */
  public function login($realm, Request $request) {
    // Initialize the realm.
    try {
      $this->session->initRealm($realm);
      $plugin = $this->session->getRealmPlugin();
    }
    catch (PluginNotFoundException $ex) {
      $this->session->destroy();
      throw new NotFoundHttpException();
    }

    // Ensure it's enabled.
    if (!$plugin->isEnabled()) {
      $this->session->destroy();
      throw new NotFoundHttpException();
    }

    // Initialize the state.
    $destination = $request->query->get('destination');
    $state = $this->session->initState($destination);

    // Get and redirect to the login URL.
    $redirect_url = Url::fromRoute('oidc.openid_connect.login_redirect');
    $url = $plugin->getLoginUrl($state, $redirect_url)
      ->setAbsolute();

    return new ImmutableTrustedRedirectResponse($url);
  }

  /**
   * Redirect callback triggered after an OpenID Connect login.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to the front page.
   */
  public function loginRedirect(Request $request) {
    $query = $request->query;
    $realm = $this->session->getRealmPluginId();

    if ($query->has('error') && $query->get('error_description')) {
      // An error occurred.
      $this->getLogger('oidc')->critical('@error error occurred on @realm realm: %description', [
        '@error' => $query->get('error'),
        '@realm' => $realm,
        '%description' => $query->get('error_description'),
      ]);
    }
    elseif (!$query->has('code')) {
      // No code specified.
      $this->getLogger('oidc')->error('Code missing in redirect from @realm realm.', [
        '@realm' => $realm,
      ]);
    }
    else {
      // Get the realm plugin.
      $plugin = $this->session->getRealmPlugin();

      try {
        // Get the tokens.
        $tokens = $plugin->getJsonWebTokens($this->session->getState(), $query->get('code'));
        $this->session->setJsonWebTokens($tokens);

        // Check for an existing user.
        $authname = $tokens->getId();
        $provider = 'oidc:' . $realm;
        $account = $this->externalauth->load($authname, $provider);

        // No user found, try to link to an existing user by email.
        if (!$account && $tokens->getEmail()) {
          $accounts = $this->userStorage->loadByProperties([
            'mail' => $tokens->getEmail(),
          ]);

          /** @var \Drupal\user\UserInterface $account */
          foreach ($accounts as $account) {
            if (!$this->authmap->getAll($account->id())) {
              $this->externalauth->linkExistingAccount($authname, $provider, $account);
              $account->setPassword(NULL)->save();
              break;
            }

            $account = NULL;
          }
        }

        // Email not set or no valid user found, create a new one.
        if (!$account) {
          do {
            $name = $this->uuid->generate();
            $count = $this->userStorage->getQuery()
              ->count()
              ->condition('name', $name)
              ->execute();
          } while ($count);

          $account = $this->externalauth->register($authname, $provider, [
            'name' => $name,
          ]);
        }

        // Login.
        $this->externalauth->userLoginFinalize($account, $authname, $provider);

        // Clear the state and restore the destination.
        if (($destination = $this->session->clearState()) !== NULL) {
          $request->query->set('destination', $destination);
        }
      }
      catch (OpenidConnectLoginException $ex) {
        if ($this->currentUser()->isAnonymous()) {
          // Redirect to the destination.
          $redirect = new ImmutableTrustedRedirectResponse($ex->getDestination());
          $this->session->destroy();
        }
        else {
          // Logout before redirecting to the destination.
          $request->query->set('destination', $ex->getDestination()->getInternalPath());
          $redirect = $this->logout($request);
        }

        // Set the exception message as error.
        if (($message = $ex->getMessage()) !== '') {
          $this->messenger()->addError($message);
        }

        return $redirect;
      }
      catch (\Exception $ex) {
        $this->getLogger('oidc')->error('Error on authentication via @realm realm: %error', [
          '@realm' => $realm,
          '%error' => $ex->getMessage(),
        ]);
      }
    }

    // Destroy the session and set an error message if authentication failed.
    if ($this->currentUser()->isAnonymous()) {
      $this->session->destroy();
      $this->messenger()->addError($this->t('Authentication failed, please try again later.'));
    }

    return $this->redirect('<front>');
  }

  /**
   * Start the OpenID Connect logout.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Drupal\oidc\Routing\ImmutableTrustedRedirectResponse
   *   Redirect to the end session endpoint.
   */
  public function logout(Request $request) {
    // Get the realm plugin.
    $plugin = $this->session->getRealmPlugin();
    $plugin_id = $this->session->getRealmPluginId();

    // Get the tokens.
    $tokens = $this->session->getJsonWebTokens();

    // Log out of Drupal. This code is copied from user_logout() because that function
    // also destroys the PHP session, which breaks the OpenID Connect session.
    $user = $this->currentUser();
    $this->getLogger('user')->notice('Session closed for %name.', ['%name' => $user->getAccountName()]);
    $this->moduleHandler()->invokeAll('user_logout', [$user]);
    $user->setAccount(new AnonymousUserSession());

    // Clear and regenerate the session.
    $this->sessionManager->clear();
    $this->sessionManager->regenerate();

    // Restore the realm since the session was destroyed.
    $this->session->initRealm($plugin_id);

    // Initialize the state.
    $destination = $request->query->get('destination');
    $state = $this->session->initState($destination);

    // Get and redirect to the logout URL.
    $redirect_url = Url::fromRoute('oidc.openid_connect.logout_redirect');
    $url = $plugin->getLogoutUrl($tokens, $state, $redirect_url)
      ->setAbsolute();

    return new ImmutableTrustedRedirectResponse($url);
  }

  /**
   * Redirect callback triggered after an OpenID Connect logout.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   *   Redirect to the front page.
   */
  public function logoutRedirect(Request $request) {
    // Set a message if there are no error messages.
    if (!$this->messenger()->messagesByType(MessengerInterface::TYPE_ERROR)) {
      $this->messenger()->addStatus($this->t('You have been logged out successfully.'));
    }

    // Clear the state and restore the destination.
    if (($destination = $this->session->clearState()) !== NULL) {
      $request->query->set('destination', $destination);
    }

    $this->session->destroy();

    return $this->redirect('<front>');
  }

}
